
/*  slsnif.c
 *  Copyright (C) 2001 Yan "Warrior" Gurtovoy (ymg@azstarnet.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "slsnif.h"
#include "ascii.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

void copyright() {
    printf("\n\nSerial Line Sniffer. Version %s\n", VERSION);
    printf("\tCopyright (C) 2001 Yan \"Warrior\" Gurtovoy (ymg@azstarnet.com)\n\n");
}

void usage() {
    copyright();
    printf("Usage: slsnif [options] <port>\n\n");
    printf("REQUIRED PARAMETERS:\n");
    printf("  <port>     - serial port to use (i.e /dev/ttyS0, /dev/ttyS1, etc.)\n\n");
    printf("OPTIONS:\n");
    printf("  -h (--help)             - displays this help.\n");
    printf("  -b (--bytes)            - print number of bytes transmitted on every read.\n");
    printf("  -t (--timestamp)        - print timestamp for every transmission.\n");
    printf("  -l (--log) <logfile>    - file to store output in, defaults to stdout.\n");
    printf("  -s (--speed) <speed>    - baudrate to use, defaults to 9600 baud.\n\n");    
    printf("Example: slsnif -l log.txt -s 2400 /dev/ttyS1\n\n");
}

void fatalError(char *msg) {
    perror(msg);
    _exit(-1);
}

int setRaw(int fd, struct termios *ttystate_orig) {
/* set tty into raw mode */

    struct termios    tty_state;
	
    if (tcgetattr(fd, &tty_state) < 0) return 0;
    /* save off original settings */
    *ttystate_orig = tty_state;
    /* set raw mode */
    tty_state.c_lflag &= ~(ICANON | IEXTEN | ISIG | ECHO);
    tty_state.c_iflag &= ~(ICRNL | INPCK | ISTRIP | IXON | BRKINT);
    tty_state.c_oflag &= ~OPOST;
    tty_state.c_cflag |= CS8;	
    tty_state.c_cc[VMIN]  = 1;
    tty_state.c_cc[VTIME] = 0;
    cfsetispeed(&tty_state, tty_data.baudrate);
    cfsetospeed(&tty_state, tty_data.baudrate);	
    if (tcsetattr(fd, TCSAFLUSH, &tty_state) < 0) return 0;
    return 1;
}

void fmtData(unsigned char *in, char *out, int in_size) {
/* format data */

    char    charbuf[15];
    int     i;

    /* flush output buffer */
    out[0] = 0;
    for (i = 0; i < in_size; i++) {
        if (in[i] == 127) {
            /* it's a DEL character */
            sprintf(charbuf, "%s (%03i) ", DEL, 127);
        } else {
            if (in[i] < 33)
                /* it's a control character or space */
                sprintf(charbuf, "%s (%03i) ", ascii_chars[(int) in[i]], in[i]);
            else
                /* it's a printable character */
                sprintf(charbuf, "%c (%03i) ", in[i], in[i]);
        }
        /* put formatted data into output buffer */
        strcat(out, charbuf);
    }
}

void writeData(int in, int out, int aux, int mode) {
/* reads data from `in`, formats it, writes it to `out` and `aux`.
 * mode 0 - read from pipe
 * mode 1 - read from port
 * mode 2 - read from pty
 */

    unsigned char   buffer[BUFFSIZE];
    char            outbuf[BUFFSIZE * 16 + 1];
    int             n;
    time_t          tstamp;

    if ((n = read(in, buffer, BUFFSIZE)) < 0) {
        if (mode)
            if (errno == EIO)
                sleep(1);
            else
                perror(mode == 1 ? RPORTFAIL : RPTYFAIL);
        else
            perror(RPIPEFAIL);
    } else {
        if (n > 0) {
            if (mode) {
                write(out, buffer, n);
                write(aux, buffer, n);                    
            } else {
                /* print timestamp if necessary */
                if (tty_data.tstamp) {
                    write(out, "\n\n", 2);
                    time(&tstamp);
                    write(out, ctime(&tstamp), 24);
                } else {
                    write(out, "\n", 1);
                }
                /* print prefix */
                write(out, aux ? PORT_IN : PORT_OUT, PRFXSIZE);
                /* format data */
                fmtData(buffer, outbuf, n);
                /* print data */
                write(out, outbuf, strlen(outbuf));
                /* print total number of bytes if necessary */
                if (tty_data.dspbytes) {
                    buffer[0] = 0;
                    sprintf(buffer, "\n%s %i", TOTALBYTES, n);
                    write (out, buffer, strlen(buffer));
                }
            }
        }
    }
}

void pipeReader() {
/* get data drom pipes */

    int             maxfd;
    fd_set          read_set;

    maxfd = max(tty_data.ptypipefd[0], tty_data.portpipefd[0]);
    while (TRUE) {
        FD_ZERO(&read_set);
        FD_SET(tty_data.ptypipefd[0], &read_set);
        FD_SET(tty_data.portpipefd[0], &read_set);
        if (select(maxfd + 1, &read_set, NULL, NULL, NULL) < 0) {
            perror(SELFAIL);
            return;
        }
        if (FD_ISSET(tty_data.ptypipefd[0], &read_set))
            writeData(tty_data.ptypipefd[0], tty_data.logfd, 0, 0);
        else
            if (FD_ISSET(tty_data.portpipefd[0], &read_set))
                writeData(tty_data.portpipefd[0], tty_data.logfd, 1, 0);
    }
}

void closeAll() {
/* close all opened file descriptors */
    /* restore settings on pty */
    if (tty_data.ptyraw)
        tcsetattr(tty_data.ptyfd, TCSAFLUSH, &tty_data.ptystate_orig);
    /* close pty */
    if (tty_data.ptyfd >= 0) close(tty_data.ptyfd);
    /* restore settings on port */
    if (tty_data.portraw)
        tcsetattr(tty_data.portfd, TCSAFLUSH, &tty_data.portstate_orig);
    /* close port */
    if (tty_data.portfd >= 0) close(tty_data.portfd);
    /* close log file */
    write(tty_data.logfd, "\n", 1);
    if (tty_data.logfd != STDOUT_FILENO && tty_data.logfd >= 0)
        if ((close(tty_data.logfd)) < 0) perror(CLOSEFAIL);
    /* close write pipes */
    if (tty_data.ptypipefd[1] >= 0) close(tty_data.ptypipefd[1]);
    if (tty_data.portpipefd[1] >= 0) close(tty_data.portpipefd[1]);
    if (pid >= 0) kill(pid, SIGINT);
}

RETSIGTYPE sigintP(int sig) {
/*parent signal handler for SIGINT */
    closeAll();
    _exit(1);
}

RETSIGTYPE sigintC(int sig) {
/* child signal handler for SIGINT */
    /* close read pipes */
    if (tty_data.ptypipefd[0] >= 0) close(tty_data.ptypipefd[0]);
    if (tty_data.portpipefd[0] >= 0) close(tty_data.portpipefd[0]);    
    _exit(1);    
}

RETSIGTYPE sigchldP(int sig) {
/* signal handler for SIGCHLD */

    int status;
    
    wait(&status);
}

int main(int argc, char *argv[]) {
    
    int     i, j, maxfd, optret;
    char    *ptyName, *portName = NULL, *logName = NULL, baudstr[7];
    fd_set  rset;
    
    struct option longopts[] = {
        {"help",      0, NULL, 'h'},
        {"log",       1, NULL, 'l'},
        {"speed",     1, NULL, 's'},
        {"bytes",     0, NULL, 'b'},
        {"timestamp", 0, NULL, 't'},
        {NULL,        0, NULL, 0}
    };
    /* initialize variables */
    baudstr[0] = 0;
    tty_data.ptyfd = tty_data.portfd = tty_data.logfd = -1;
    tty_data.ptyraw = tty_data.portraw = FALSE;
    tty_data.dspbytes = tty_data.tstamp = FALSE;
    tty_data.ptypipefd[0] = tty_data.ptypipefd[1] = -1;
    tty_data.portpipefd[0] = tty_data.portpipefd[1] = -1;
    tty_data.baudrate = DEFBAUDRATE;
    /* activate signal handlers */
    signal(SIGINT, sigintP);
    signal(SIGCHLD, sigchldP);
    /* register closeAll() function to be called on normal termination */
    atexit(closeAll);
    /* process command line arguments */
    while ((optret = getopt_long(argc, argv, OPTSTR, longopts, NULL)) != -1) {
        switch (optret) {
            case 'b':
                tty_data.dspbytes = TRUE;
                break;
            case 't':
                tty_data.tstamp = TRUE;
                break;
            case 'l':
                logName = optarg;
                break;
            case 's':
                i = 0;
                while (baudrates[i].spdstr &&
                        strcmp(optarg, baudrates[i].spdstr)) i++;
                if (baudrates[i].spdstr) {
                    tty_data.baudrate = baudrates[i].speed;
                    strcat(baudstr, baudrates[i].spdstr);
                }
                break;
            case 'h':
            case '?':
            default :
                usage();
                return -1;
        }
    }
    if (optind < argc) portName = argv[optind];
    if (!portName) {
        usage();
        return -1;
    }
    copyright();
    /* open logfile */
    if (!logName)
        tty_data.logfd = STDOUT_FILENO;
    else {
        if((tty_data.logfd = open(logName,
                    O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR)) < 0) {
            perror(LOGFAIL);
            tty_data.logfd = STDOUT_FILENO;
        } else {
            printf("Started logging data into file '%s'.\n", logName);
        }
    }
    /* create pipe */
    if (pipe(tty_data.ptypipefd) < 0 || pipe(tty_data.portpipefd) < 0) {
        perror(PIPEFAIL);
        return -1;
    }
    /* fork child process */
    switch (pid = fork()) {
    case 0:
        /* child process */
        /* close write pipe */
        close(tty_data.ptypipefd[1]);
        close(tty_data.portpipefd[1]);
        signal(SIGINT, sigintC);
        pipeReader(&tty_data);
        break;
    case -1:
        /* fork() failed */
        perror(FORKFAIL);
        return -1;
    default:
        /* parent process */
        /* close read pipe */
        close(tty_data.ptypipefd[0]);
        close(tty_data.portpipefd[0]);
        break;
    }
    /* open master side of the pty */
    /* Search for a free pty */
    if (!(ptyName = strdup(DEFPTRNAME))) fatalError(MEMFAIL);
    for(i = 0; i < 16 && tty_data.ptyfd < 0; i++) {
        ptyName[8] = "pqrstuvwxyzPQRST"[i];
        for(j = 0; j < 16 && tty_data.ptyfd < 0; i++) {
            ptyName[9] = "0123456789abcdef"[j];
            /* try to open master side */
            tty_data.ptyfd = open(ptyName, O_RDWR|O_NONBLOCK|O_NOCTTY);
        } 
    }
    if (tty_data.ptyfd < 0) {
        /* failed to find an available pty */
        free(ptyName);
        perror(PTYFAIL);
        return -1;
    }
    /* create the name of the slave pty */
    ptyName[5] = 't';
    printf("Opened pty: %s\n", ptyName);
    free(ptyName);
    /* set raw mode on pty */
    if(!setRaw(tty_data.ptyfd, &tty_data.ptystate_orig)) {
        perror(RAWFAIL);
        return -1;
    }
    tty_data.ptyraw = TRUE;
    /* try to open port */
    if ((tty_data.portfd = open(portName, O_RDWR|O_NONBLOCK)) < 0) {
        perror(PORTFAIL);
        return -1;
    }
    printf("Opened port: %s\n", portName);
    /* set raw mode on port */
    if (!setRaw(tty_data.portfd, &tty_data.portstate_orig)) {
        perror(RAWFAIL);
        return -1;
    }
    tty_data.portraw = TRUE;
    printf("Baudrate is set to %s baud%s.\n",
                baudstr[0] ? baudstr : "9600",
                baudstr[0] ? "" : " (default)");
    /* start listening to the slave and port */
    maxfd = max(tty_data.ptyfd, tty_data.portfd);
    while (TRUE) {
        FD_ZERO(&rset);
        FD_SET(tty_data.ptyfd, &rset);
        FD_SET(tty_data.portfd, &rset);
        if (select(maxfd + 1, &rset, NULL, NULL, NULL) < 0) {
            perror(SELFAIL);
            return -1;
        }
        if (FD_ISSET(tty_data.portfd, &rset)) {
            writeData(tty_data.portfd, tty_data.ptyfd, tty_data.portpipefd[1], 1);
        } else {
            if (FD_ISSET(tty_data.ptyfd, &rset)) {
                writeData(tty_data.ptyfd, tty_data.portfd, tty_data.ptypipefd[1], 2);
            }
        }
    }
    return 1;
}